from dataclasses import dataclass
from typing import Optional

from edps import subworkflow, task, Task

from .eris_datasources import *
from .eris_task_function import is_pupil_tracking, not_pupil_tracking
from .eris_task_function import is_short, is_short_broad, is_standard_star
from .eris_task_function import which_band, which_target, which_nix_mode


@dataclass
class NixProcessedObservations:
    science_reduction: Task
    sky_subtraction: Optional[Task] = None
    astrometry: Optional[Task] = None
    photometry: Optional[Task] = None
    stacking: Optional[Task] = None


def nix_science(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight, observ_mode, observ_type,
                metatargets):
    task_building = (task('process_nix_' + observ_type + '_' + observ_mode)
                     .with_main_input(raw_inputs)
                     .with_dynamic_parameter("band_used", which_band)
                     .with_dynamic_parameter("std_or_science", which_target)
                     .with_dynamic_parameter("nix_mode", which_nix_mode)
                     .with_associated_input(linearity_nix, [GAIN_INFO, COEFFS_CUBE, BP_MAP_NL])
                     .with_associated_input(dark_nix, [MASTER_DARK_IMG])
                     .with_associated_input(nix_wcs_matched_catalogue)
                     .with_associated_input(nix_wcs_refine)
                     .with_associated_input(nix_phot_data, condition=is_standard_star)
                     .with_meta_targets(metatargets))

    # flats are associated as alternative to coronography and sam observations.
    # These modes are not supported by a recipe
    if observ_mode in ["coronography", "sam"]:
        task_building = (task_building.with_associated_input(nix_flat_calibrations,
                                                             [MASTER_FLAT_LAMP_HIFREQ, MASTER_FLAT_LAMP_LOFREQ,
                                                              MASTER_BPM_LAMP], min_ret=0))
    else:
        # flats are associated as alternative associated inputs to ifu, images, and long-slit.
        task_building = (task_building.with_alternative_associated_inputs(nix_flat_calibrations))
        # twilight flats are associated only for observing modes different from "coronography" and "sam"
        task_building = (task_building.with_associated_input(nix_flat_twilight, [MASTER_FLAT_TWILIGHT_LOFREQ],
                                                             condition=is_short_broad,
                                                             match_rules=match_twflat_short_broad)
                         .with_associated_input(nix_flat_twilight, [MASTER_FLAT_TWILIGHT_LOFREQ],
                                                condition=is_short, match_rules=match_twflat_short))

        # These modes are supported by a recipe
        task_building = (task_building.with_recipe('eris_nix_cal_det'))
    if observ_mode == "images":
        task_building = (task_building.with_associated_input(raw_sky_jitter_image, min_ret=0, max_ret=100))
    if observ_mode == 'long_slit':
        task_building = (task_building.with_associated_input(raw_sky_jitter_lss, min_ret=0, max_ret=100))
    if observ_mode == "cube":
        task_building = (task_building.with_associated_input(raw_sky_jitter_cube, min_ret=0, max_ret=100))

    return task_building.build()


def nix_reduction(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight, observ_mode,
                  observ_type, metatargets):
    # This subworkflow processes on-sky NIX exposures (science, standard, and astrometric).
    # The input variable observ_mode specifies whether the raw_inputs are coronography, images, ifu or
    # long-slit spectra. It is used to define the reduction cascade and the return arguments.
    # The input variable obs_type specifies whether the input data are science, standar, or astrometry.
    # Task names are set accordingly to observ_type and observ_mode.

    # Coronography and SAM exposures are not yet supported by the pipeline.
    # Thus, they are used only for visual inspection.
    # CUBE-format exposures are supported by the pipeline, and the instrument signature are removed from the data.
    if observ_mode in ["cube", "coronography", "sam"]:
        return nix_science(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight, observ_mode,
                           observ_type, metatargets)

    # Reductio of LONG-SLIT and IMAGES exposures continue
    # This tasks removes the instrument signature from the data. It is executed for all types of inputs (images, ifu,
    # or long-slit spectra)
    science = nix_science(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight, observ_mode,
                          observ_type, metatargets)

    # IMAGES and LONG-SLIT observations are processed further
    sky_subtraction = (task('sky_subtraction_nix_' + observ_type + '_' + observ_mode)
                       .with_recipe('eris_nix_img_skysub')
                       .with_main_input(science)
                       .with_meta_targets(metatargets)
                       .build())

    astrometry = (task('astrometry_nix_' + observ_type + '_' + observ_mode)
                  .with_recipe('eris_nix_img_cal_wcs')
                  .with_main_input(sky_subtraction)
                  .with_associated_input(nix_wcs_matched_catalogue)
                  .with_meta_targets(metatargets)
                  .build())

    # Only images observations generate IDPs.
    photometry = (task('photometry_nix_' + observ_type + '_' + observ_mode)
                  .with_recipe('eris_nix_img_cal_phot')
                  .with_main_input(astrometry)
                  .with_associated_input(nix_phot_data)
                  .with_min_group_size(2)
                  .with_meta_targets(metatargets)
                  .build())

    # Only science image observations are stacked
    if observ_type == "science" and observ_mode == "images":
        stacking = (task('combine_nix_' + observ_type + '_' + observ_mode)
                    .with_condition(not_pupil_tracking)
                    .with_dynamic_parameter("is_pupil_tracking", is_pupil_tracking)
                    .with_recipe('eris_nix_img_hdrl_stack')
                    .with_main_input(photometry)
                    .with_meta_targets(metatargets)
                    .build())

        return NixProcessedObservations(science_reduction=science,
                                        sky_subtraction=sky_subtraction,
                                        astrometry=astrometry,
                                        photometry=photometry,
                                        stacking=stacking)

    else:  # long slit or non-science exposure stop without the combination.
        return NixProcessedObservations(science_reduction=science,
                                        sky_subtraction=sky_subtraction,
                                        astrometry=astrometry,
                                        photometry=photometry)


@subworkflow("science_reduction_nix", "")
def nix_science_reduction(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight, observ_mode,
                          observ_type, metatargets):
    return nix_reduction(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight, observ_mode,
                         observ_type, metatargets)


@subworkflow("on_sky_calibrations_reduction_nix", "")
def nix_on_sky_calibrations_reduction(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight,
                                      observ_mode, observ_type, metatargets):
    return nix_reduction(raw_inputs, linearity_nix, dark_nix, nix_flat_calibrations, nix_flat_twilight, observ_mode,
                         observ_type, metatargets)
